/*
	Plugin: Roundend Blocker
	Current Version: 1.1.065
	
	Author: Nextra
	E-Mail: nextra.24@gmail.com
	
	Support-Thread: http://forums.alliedmods.net/showthread.php?t=85365
	
		AMX Mod X script.
		
		This program is free software; you can redistribute it and/or modify it
		under the terms of the GNU General Public License as published by the
		Free Software Foundation; either version 2 of the License, or (at
		your option) any later version.

		This program is distributed in the hope that it will be useful, but
		WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
		General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with this program; if not, write to the Free Software Foundation,
		Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

		In addition, as a special exception, the author gives permission to
		link the code of this program with the Half-Life Game Engine ("HL
		Engine") and Modified Game Libraries ("MODs") developed by Valve,
		L.L.C ("Valve"). You must obey the GNU General Public License in all
		respects for all of the code used other than the HL Engine and MODs
		from Valve. If you modify this file, you may extend this exception
		to your version of the file, but you are not obligated to do so. If
		you do not wish to do so, delete this exception statement from your
		version.

		
	.: Description
	
		This plugin creates fake bots in both teams to prevent the ending of rounds 
		normally occuring because every player on the map died. 
		It is only useful in a deathmatch environment where all players respawn infinitely.

	.: Features

		- Bots automatically join the game when only a few players are on the map
		- The bots automatically leave the game as more players join the game
		- Bots are shown as spectators on the scoreboard, and are far away from the battlefield 
			(and also invisible and not solid) so they don't cause confusion and can't be killed
		- Fakefull mode. 
			In Fakefull mode bots will always be on the server. They will connect on startup when 
			no players are online and will only leave when the playercount reaches plrmax. 
			If this mode is disabled the bots will not connect on startup and will disconnect 	
			when they are not needed anymore (all players disconnected).

	.: Cvars

		reb_enable <0|1> - 0 = Disable, 1 = Enable the plugin
		reb_plrmax - When the playercount (only players in a team) reaches this value, kick the bots
		reb_plrmin - When the playercount goes below this value, add the bots
		reb_fakefull <0|1> - 0 = Disable, 1 = Enable fakefull mode
		reb_fullkick <0|1> - 0 = Bots stay on the server even if full, 1 = Last slot will stay free, bots will be kicked
		reb_name1 "" - Sets the name for the first bot
		reb_name2 "" - Sets the name for the second bot

	.: Required Modules

		- CStrike
		- FakeMeta
		- Ham Sandwich

	.: Credits
	
		- Thanks to jim_yang, he created the original roundend_blocker on which my version is based on. 
			He allowed me to submit my recode of this to the public
			
	.: Notes
	
		- Due to a recent change the bots will no longer reconnect if there is no player on the server and all bots 
			are kicked in rapid succession. I am assuming that doing so requires the intention of having no bots on the 
			server so I do not consider this an issue I have to fix. It is more important to have no crashes around.

	.: Changelog
	
		*	1.1.045 
		- Initial release.
		*	1.1.050
		- Changed sequence of checks on preThink for better efficiency.
		- Changed adding routine to avoid problems with a high BOT_NUM (or bots filling up the server to the limit somehow).
		* 	1.1.055
		- Changed hiding routine to allow hiding of single bots after creation.
		- Fixed possible error/crash when BOT_NUM was higher than maxslots.
		- Added public cvar.
		*	1.1.060
		- Added fakefull mode.
		- Fixed creation routine to prevent recursion when plrmin equals plrmax and this playercount is currently on the server.
		- Changed creation routine to prevent unwanted bot creations between changelevel and plugin_end, fixing crashes
			that occured in some situations.
		*	1.1.065
		- Replaced preThink hook by TeamInfo message hook for better efficiency. Thanks to Empowers.
		- Added fullkick cvar to make all bots stay on the server even if it's full. Suggestion by Empowers.
*/

#include <amxmodx>
#include <amxmisc>
#include <cstrike>
#include <fakemeta>
#include <hamsandwich>

#pragma semicolon 1
#pragma ctrlchar '\'

// Some constants you may want to change

const BOT_NUM = 2;	// If you, for some reason, want more bots.
					// Change this and don't forget to add reb_name3, reb_name4 and so on

// These are just the default botnames, they are only used if the cvar is not set! You don't need to change them!
// You do, however, have to add more to this list if you change BOT_NUM
new const DefaultNames[BOT_NUM][ ] = {
	"Have Fun",
	"Good Luck"
};

//----------------------------------------------------------------------------------------
// Do not change anything below this line if you don't know exactly what you are doing!
//----------------------------------------------------------------------------------------
new const PLUGIN[] 	= "Roundend Blocker";
new const VERSION[] = "1.1.065";
new const AUTHOR[] 	= "Nextra";

new p_BlockEnd, p_BlockPlrMin, p_BlockPlrMax, p_FakeFull, p_FullKick, p_BlockName[BOT_NUM];

const TASK_BOTS = 666;

new g_BotID[BOT_NUM], bool:g_bFirstRound = true;

new g_iMaxPlayers;

public plugin_init( )
{
	register_plugin( PLUGIN, VERSION, AUTHOR );
	register_cvar( "reb_version", VERSION, FCVAR_SERVER|FCVAR_SPONLY );

	register_logevent( "on_RoundEnd"	, 2, "1=Round_End" 			);
	register_logevent( "on_RoundStart"	, 2, "1=Round_Start"		);

	register_message( get_user_msgid( "TeamInfo" ), "msg_TeamInfo" 	);

	p_BlockEnd		= register_cvar( "reb_enable"	,	"1" 		),
	p_BlockPlrMax	= register_cvar( "reb_plrmax"	,	"10" 		),
	p_BlockPlrMin	= register_cvar( "reb_plrmin"	,	"8"			),
	p_FakeFull		= register_cvar( "reb_fakefull"	,	"1"			),
	p_FullKick		= register_cvar( "reb_fullkick" ,	"1"			);
	
	new szTmp[16];
	
	for( new i = 0; i < BOT_NUM; i++ )
	{
		formatex( szTmp, charsmax(szTmp), "reb_name%i", i + 1 );
		
		p_BlockName[i] = register_cvar( szTmp, "" );
	}

	g_iMaxPlayers 		= get_maxplayers( );
}


public msg_TeamInfo( )
{	
	if( is_user_fakebot( get_msg_arg_int( 1 ) ) )
		set_msg_arg_string( 2, "SPECTATOR" );
}


public on_RoundEnd( )
	BOT_start( );


public on_RoundStart( )
{
	if( g_bFirstRound )
	{
		g_bFirstRound = false;
		
		if( get_pcvar_num( p_FakeFull ) && get_pcvar_num( p_BlockEnd ) )
		{
			if( !get_playersnum( 1 ) )
				set_task( 1.0, "BOT_create", TASK_BOTS );
		}
	}
	else if( BOT_num( ) )
		BOT_hide( );
}


public BOT_create( )
{
	new BotName[32], BotID, firstNotValid = -1, i;
	
	for( i = 0; i < BOT_NUM; i++ )
	{
		if( !IsValidPlayer( g_BotID[i] ) )
		{
			firstNotValid = i;
			break;
		}
	}
	
	if( firstNotValid == -1 )
		return;

	get_pcvar_string( p_BlockName[firstNotValid], BotName, charsmax(BotName) );
	trim( BotName );
	
	if( equali( BotName, "" ) )
		copy( BotName, charsmax(BotName), DefaultNames[firstNotValid] );
	
	trim( BotName );
	
	if( equali( BotName, "" ) )
		return;

	new szName[32], szTmp[32], bool:bContinue = true, bool:bFound, iFound;
	
	copy( szTmp, charsmax(szTmp), BotName );
	
	while( bContinue )
	{
		for( i = 1; i <= g_iMaxPlayers; i++ )
		{
			if( is_user_connected( i ) )
			{
				get_user_name( i, szName, charsmax(szName) );
				
				if( equali( szTmp, szName ) )
				{
					formatex( szTmp, charsmax(szTmp), "%s(%d)", BotName, ++iFound );
					
					bFound = true;
					
					break;
				}
			}
		}
		
		if( !bFound )
			bContinue = false;
		
		bFound = false;
	}
	
	if( iFound )
		copy( BotName, charsmax(BotName), szTmp );
	
	BotID = engfunc( EngFunc_CreateFakeClient, BotName );
	
	if( !IsValidPlayer( BotID ) )
		return;	

	g_BotID[firstNotValid] = BotID;
	
	new ptr[128];
	engfunc( EngFunc_FreeEntPrivateData, BotID );
	dllfunc( DLLFunc_ClientConnect, BotID, BotName, "127.0.0.1", ptr );
	
	if( !is_user_connected( BotID ) )
	{
		g_BotID[firstNotValid] = 0;
		return;	
	}

	dllfunc( DLLFunc_ClientPutInServer, BotID );
	set_pev( BotID, pev_spawnflags, pev( BotID, pev_spawnflags ) | FL_FAKECLIENT );
	set_pev( BotID, pev_flags, pev( BotID, pev_flags ) | FL_FAKECLIENT );
	cs_set_user_team( BotID, BOT_num( ) % 2 ? CS_TEAM_T : CS_TEAM_CT );

	if( !is_user_alive( BotID ) )
		ExecuteHamB( Ham_CS_RoundRespawn, BotID );
	
	if( BOT_num( ) < BOT_NUM )
	{
		remove_task( TASK_BOTS );
		set_task( 0.1, "BOT_start", TASK_BOTS );
	}
	
	BOT_hide( firstNotValid );
}


BOT_hide( which = BOT_NUM )
{
	new BotID;

	switch( which )
	{
		case BOT_NUM:
		{
			for( new i = 0; i < BOT_NUM; i++ ) 
			{
				BotID = g_BotID[i];
				
				if( IsValidPlayer( BotID ) )
				{
					set_pev( BotID, pev_effects, pev( BotID, pev_effects ) | EF_NODRAW );
					set_pev( BotID, pev_solid, SOLID_NOT );
					set_pev( BotID, pev_takedamage, DAMAGE_NO );
					set_pev( BotID, pev_origin, Float:{9999.0,9999.0,9999.0} );
				}
			}
		}
		default:
		{
			BotID = g_BotID[which];
			
			if( IsValidPlayer( BotID ) )
			{
				set_pev( BotID, pev_effects, pev( BotID, pev_effects ) | EF_NODRAW );
				set_pev( BotID, pev_solid, SOLID_NOT );
				set_pev( BotID, pev_takedamage, DAMAGE_NO );
				set_pev( BotID, pev_origin, Float:{9999.0,9999.0,9999.0} );
			}
		}
	}
}



BOT_kick( iAmt = BOT_NUM )
{
	if( !BOT_num( ) )
		return;
	
	new BotID, iKicked;
	
	for( new i = BOT_NUM - 1; i >= 0; i-- )
	{
		if( i + 1 > g_iMaxPlayers )
			continue;
		
		BotID = g_BotID[i];
		
		if( IsValidPlayer( BotID ) )
		{
			server_cmd( "kick #%d", get_user_userid( BotID ) );
			server_exec( );
			
			if( ++iKicked >= iAmt )
				break;
		}
	}
}


public BOT_start( )
{
	new iBotNum = BOT_num( );
	
	if( !get_pcvar_num( p_BlockEnd ) )
	{
		if( iBotNum )
			BOT_kick( );
		
		return;
	}
	
	new iPlayers = get_playersnum( 1 );

	if( iBotNum )
	{
		if( iPlayers == g_iMaxPlayers - 1 )
		{
			if( get_pcvar_num( p_FullKick ) )
				return;
		}
		else if( iPlayers >= g_iMaxPlayers )
		{
			if( get_pcvar_num( p_FullKick ) )
				BOT_kick( 1 );
			
			return;
		}
		else if( iPlayers - iBotNum == 0 )
		{
			if( !get_pcvar_num( p_FakeFull ) )
			{
				BOT_kick( );
				return;
			}
		}
	}
	else if( !iPlayers )
		return;
		
	new iPlayerNum;
	
	for( new i = 1; i <= g_iMaxPlayers; i++ )
	{
		if( !is_user_connected( i ) || is_user_fakebot( i ) )
			continue;
		
		switch( cs_get_user_team( i ) )
		{
			case CS_TEAM_T, CS_TEAM_CT:	iPlayerNum++;
		}
	}
	
	
	if( iBotNum && iPlayerNum >= get_pcvar_num( p_BlockPlrMax ) )
		BOT_kick( );
	else if( iBotNum < BOT_NUM && iPlayerNum < get_pcvar_num( p_BlockPlrMin ) )
		BOT_create( );
}


BOT_num( )
{
	new iNum;
	
	for( new i = 0; i < BOT_NUM; i++ )
	{
		if( IsValidPlayer( g_BotID[i] ) )
			iNum++;
	}
	
	return iNum;
}


public client_putinserver( id )
	set_task( 1.0, "BOT_start", TASK_BOTS );


public client_disconnect( id )
{	
	for( new i = 0; i < BOT_NUM; i++ )
	{
		if( is_user_fakebot( id, i ) )
		{
			g_BotID[i] = 0;
			
			break;
		}
	}
	
	set_task( 1.0, "BOT_start", TASK_BOTS );
}


bool:IsValidPlayer( const id )
	return ( 1 <= id <= g_iMaxPlayers ) ? true : false;


bool:is_user_fakebot( const id, which = BOT_NUM )
{
	switch( which )
	{
		case BOT_NUM:
		{
			for( new i = 0; i < BOT_NUM; i++ )
			{
				if( id == g_BotID[i] )
					return true;
			}
		}
		default: return ( id == g_BotID[which] ) ? true : false;
	}
	
	return false;
}